/*
 * Basic Windows entry point and config engine.
 * Part of Storming Ions (a screensaver).
 * Copyright (C) 2007 W. Ross Younger.
 * See the included file "COPYING" for licensing information.
 *
 * Portions of this file derive from work in Simon Tatham's Portable Puzzle
 * Collection, so may be subject to his copyright. I assert that this is in
 * compliance with that software's copyright and permissions message, and that
 * they do not constitute a "substantial portion" of that software, so there 
 * is no need to reproduce that message.
 */

#define _CRT_SECURE_NO_DEPRECATE /* for the vsnprintf warning */

#include <windows.h>
#include <scrnsave.h>
#include <time.h>
#include <stdio.h>

#include "base.h"
#include "gfx.h"
#include "config.h"

/* ******************************************************************** */

void error_msg(const char *fmt, ...) {
	char buf[4096];
	va_list va;

	va_start(va, fmt);
	_vsnprintf(buf, 4095, fmt, va);
	va_end(va);
	MessageBox(0, buf, "Error", MB_OK | MB_ICONERROR);
}

LONG WINAPI ScreenSaverProc(HWND hwnd, UINT message, WPARAM wParam, LONG lParam) 
{ 
	static HDC          hdc;    /* device-context handle */ 
	static RECT         rect;     /* RECT structure */ 
	static UINT         uTimer; /* timer identifier */      

	int result;
	int xPos,yPos;
	int redraw;

	switch(message)     
	{
        case WM_CREATE:
			hdc = GetDC(hwnd);             
			GetClientRect(hwnd, &rect);
			redraw = config_get("Redraw interval");

			result=ss_setup(hdc, rect);
			ReleaseDC(hwnd,hdc);
			if (result<0) return result;
			
			// callback timer - milliseconds
			uTimer = SetTimer(hwnd, 1, redraw, NULL);  

		 break; 

         case WM_ERASEBKGND:             
			/* From the MSDN notes:
				WM_ERASEBKGND is called once before the first WM_TIMER
				-- we should blank the background at that point */
            hdc = GetDC(hwnd);             
			GetClientRect (hwnd, &rect); 
            FillRect (hdc, &rect, (HBRUSH) GetStockObject(BLACK_BRUSH)); 
            ReleaseDC(hwnd,hdc);             
			break;
		 
		 case WM_TIMER:  
			/* Will be called regularly by the system */
			hdc = GetDC(hwnd);
			GetClientRect(hwnd, &rect); 
			ss_frame(hdc,rect);
			ReleaseDC(hwnd,hdc);
			break;  

        case WM_DESTROY: 
			if (uTimer) KillTimer(hwnd, uTimer);
			ss_cleanup();
			break;   
			
		case WM_MOUSEMOVE:
			xPos = LOWORD(lParam);
			yPos = HIWORD(lParam);
			// If desired, do something with this.
			// Return TRUE to trap the message and prevent wake-up.
	}
	
	/* DefScreenSaverProc processes things we haven't */
    return DefScreenSaverProc(hwnd, message, wParam, lParam); 
}  

/* No child controls are required. */
BOOL WINAPI RegisterDialogClasses (HANDLE hInst)
{
	return TRUE;
}



/* ScreenSaverConfigureDialog

Called from the SS Library to handle dialog events.
(Note to the unwary: this is called iff there's a resource present of the correct ID.)
See: http://msdn2.microsoft.com/en-us/library/bb776796.aspx
*/

config_item **conf = 0;
int id_defaults;

extern HINSTANCE hMainInstance;
char version_string[42];

void load_strings(void) {
	LoadString(hMainInstance, VERSION_STRING_ID, version_string, sizeof(version_string)/sizeof(*version_string));
}

/* Drawing heavily on windows.c in Simon Tatham's portable puzzles collection */
void draw_dialog(HWND dlg) {
	int height, width;
	TEXTMETRIC tm;
	SIZE size;
	HDC dc = GetDC(dlg);
	HGDIOBJ font, oldfont;

	conf=configs();

	font = CreateFont(-MulDiv(8, GetDeviceCaps(dc, LOGPIXELSY), 72),
                             0, 0, 0, 0,
                             FALSE, FALSE, FALSE, DEFAULT_CHARSET,
                             OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                             DEFAULT_QUALITY,
                             FF_SWISS,
                             "MS Shell Dlg");

	oldfont = SelectObject(dc, font);

	/* How big does this dialog have to be? */
	if (GetTextMetrics(dc, &tm)) {
		height = tm.tmAscent + tm.tmDescent;
		width = tm.tmAveCharWidth;
	}
	else
		height = width = 30;

	int maxlabel=0, maxcheckbox=0, winheight = height/2;
	
	for (unsigned i=0; i<n_configdefs; i++) {
		if (!GetTextExtentPoint32(dc, conf[i]->name, strlen(conf[i]->name), &size)) {
			error_msg("Couldn't figure out size of a text string (huh?) (%d)", GetLastError());
			continue;
		}
		switch(conf[i]->type) {
			case T_BOOL:
				if (maxcheckbox < size.cx)
					maxcheckbox = size.cx;
				winheight += height*3/2;
				break;

			case T_INT:
			case T_FLOAT:
			case T_SELECT:
				if (GetTextExtentPoint32(dc, conf[i]->help, strlen(conf[i]->help), &size) &&
						maxlabel < size.cx)
					maxlabel = size.cx;
				if (conf[i]->type == T_SELECT)
					winheight += height*2;
				else
					winheight += height*9/2;
				break;
		}
	}

	winheight += height*23/4;

	int col1l, col1r, col2l, col2r;
    col1l = 2*width;
    col1r = col1l + maxlabel;
    col2l = col1r + 2*width;
    col2r = col2l + 30*width;
    if (col2r < col1l+2*height+maxcheckbox)
    col2r = col1l+2*height+maxcheckbox;
    int winwidth = col2r + 2*width;

	/* Now we have height and width, resize the dialog */
	if (!SetWindowPos(dlg, HWND_TOP, 0,0, winwidth, winheight, SWP_NOZORDER | SWP_NOMOVE))
		error_msg("Could not SetWindowPos: %d", GetLastError());

	SendMessage(dlg, WM_SETFONT, (WPARAM) font, FALSE);

	/* Hope this works. Now we can possibly get on with setting up the controls. */
	int id = 1000, y=0;
	HWND tmp;

	unsigned slen = strlen(version_string)+1;
	tmp=CreateWindowEx(0, "Static", version_string, WS_CHILD | WS_VISIBLE,
		col2r-width*slen, y, width*slen, height, dlg, (HMENU)(id++), 0, 0);
	SendMessage(tmp, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));
	y+=height;

	for (unsigned i=0; i<n_configdefs; i++) {
		switch (conf[i]->type) {
			case T_BOOL: // checkbox
				tmp=CreateWindowEx(0, "Button", conf[i]->help, 
					BS_NOTIFY | BS_AUTOCHECKBOX | WS_TABSTOP | WS_CHILD | WS_VISIBLE,
					col1l, y+height/4, col2r-col1l, height, dlg, (HMENU)(conf[i]->ctlid = (id++)), 0, 0);
				SendMessage(tmp, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));
				CheckDlgButton(dlg, conf[i]->ctlid, conf[i]->val);
				y += height*3/2;
	        break;

			case T_SELECT: // labelled drop-down list
				tmp=CreateWindowEx(0, "Static", conf[i]->help, WS_CHILD | WS_VISIBLE,
					col1l, y+height/8, col1r-col1l, height, dlg, (HMENU)(id++), 0 /*...*/, 0);
				SendMessage(tmp, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));
				{
					HWND ctl = CreateWindowEx(WS_EX_CLIENTEDGE, "ComboBox", 0, 
						WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS | WS_CHILD | WS_VISIBLE,
						col2l, y, col2r-col2l, height*41/2, dlg, (HMENU)(conf[i]->ctlid = (id++)), 0, 0);
					SendMessage(ctl, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));
					SendMessage(ctl, CB_RESETCONTENT, 0, 0);
					const keyval **kv = conf[i]->_kv;
					int idx = 0, defidx;
					while ((*kv)->key) {
						SendMessage(ctl, CB_ADDSTRING, 0, (LPARAM)((*kv)->key));
						if ((*kv)->val == conf[i]->val) defidx = idx;
						++kv, ++idx;
					}
					SendMessage(ctl, CB_SETCURSEL, defidx, 0);
				}
				y += height*2;
				break;

			case T_INT:
			case T_FLOAT:
				/* OK so we need...
				 * a scrollbar (whose ID = this ctlid)
				 * a string to put the "current" value (ID = this ctlid + 1, eep)
				 * possibly a label at either end (leave off for now)
				 * a titled GROUPBOX surrounding the whole shebang */
				{
					int baseX = col1l, baseY = y+height/8;
					HWND gbox = CreateWindowEx(0, "BUTTON", conf[i]->help, BS_GROUPBOX | WS_CHILD | WS_VISIBLE,
						baseX, baseY, col2r-col1l, height*4, 
						dlg, (HMENU)(id++), 0, 0);
					SendMessage(gbox, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));

					HWND sbar = CreateWindowEx(0, "ScrollBar", "foo", WS_CHILD | WS_VISIBLE,
						baseX+width, baseY+9*height/4, (col2r-col1l)-2*width, 3*height/2,
						dlg, (HMENU)(conf[i]->ctlid = (id++)), 0, 0);
					int midpt = ( (col2r-col1l)-2*width ) / 2;
					tmp=CreateWindowEx(0, "Static", "" /* will be updated later */, WS_CHILD | WS_VISIBLE,
						baseX+midpt, baseY+height, midpt, height, 
						dlg, (HMENU)(id++), 0, 0);
					SendMessage(tmp, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));
					/* so ctlid = scrollbar; ctlid+1 = roving value legend */

					/* sbar = GetDlgItem(dlg, id); */
					SCROLLINFO si;
					si.cbSize = sizeof(si);
					si.nMin = conf[i]->_min;
					si.nMax = conf[i]->_max;
					si.nPos = conf[i]->_default;
					si.nPage = conf[i]->_step;
					si.fMask = SIF_ALL;
					SetScrollInfo(sbar, SB_CTL, &si, TRUE);
				}
				y += height*9/2;
				break;
			default:
				error_msg("Unhandled config type %d",conf[i]->type);
		}
	}

	y += height/2;
	int third = winwidth/3;
	int bwidth = third - width*2;
	tmp = CreateWindowEx(0, "Button", "OK", BS_PUSHBUTTON | WS_TABSTOP | BS_DEFPUSHBUTTON | WS_CHILD | WS_VISIBLE,
		width, y, bwidth, height*7/4, dlg, (HMENU)IDOK, 0, 0);
	SendMessage(tmp, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));

	tmp = CreateWindowEx(0, "Button", "Cancel", BS_PUSHBUTTON | WS_TABSTOP | WS_CHILD | WS_VISIBLE,
		third+width, y, bwidth, height*7/4, dlg, (HMENU)IDCANCEL, 0, 0);
	SendMessage(tmp, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));

	tmp = CreateWindowEx(0, "Button", "Defaults", BS_PUSHBUTTON | WS_TABSTOP | WS_CHILD | WS_VISIBLE,
		third*2+width, y, bwidth, height*7/4, dlg, (HMENU)(id_defaults = id++), 0, 0);
	SendMessage(tmp, WM_SETFONT, (WPARAM)font, MAKELPARAM(TRUE,0));

	/* Right! Finally! */
	SelectObject(dc,oldfont);
	ReleaseDC(dlg, dc);
	EnableWindow(dlg,TRUE);
}


void redraw_dialog(HWND dlg) {
	for (unsigned i=0; i<n_configdefs; i++) {
		switch(conf[i]->type) {
			case T_BOOL:
				CheckDlgButton(dlg, conf[i]->ctlid, conf[i]->val == 0 ? BST_UNCHECKED: BST_CHECKED);
				break;

			case T_INT:
			case T_FLOAT:
			{
				HWND sbar = GetDlgItem(dlg, conf[i]->ctlid);
				SCROLLINFO si;
				si.cbSize = sizeof(si);
				si.nPos = conf[i]->val;
				si.fMask = SIF_POS;
				SetScrollInfo(sbar, SB_CTL, &si, TRUE);

				if (conf[i]->type == T_INT) {
					SetDlgItemInt(dlg, conf[i]->ctlid+1, conf[i]->val, FALSE);
				} else {
					char buf[80];
					sprintf_s(buf, sizeof(buf), "%d/%d", (conf[i]->val), conf[i]->_denom);
					buf[sizeof(buf)-1]=0;
					SetDlgItemText(dlg, conf[i]->ctlid+1, buf);
				}
				break;
			}

			case T_SELECT:
				break; /* Windows does it for us ... */
		}
	}
}

BOOL WINAPI ScreenSaverConfigureDialog(HWND hDlg, UINT message, 
									   WPARAM wParam, LONG lParam) 
{
	int i;
	LPARAM p = LOWORD(wParam);
	switch(message) {
		case WM_INITDIALOG:
			load_strings();
			draw_dialog(hDlg);
			redraw_dialog(hDlg);
			return TRUE;

		case WM_HSCROLL:
			for (i=0; i<n_configdefs; i++) {
				config_item *c = conf[i];
				if ( (c->type != T_INT) && (c->type != T_FLOAT) ) 
					continue;
				if (GetDlgItem(hDlg, c->ctlid) != (HWND)lParam)
					continue;
				
				switch (LOWORD(wParam)) { 
		            case SB_PAGELEFT:  c->val -= c->_step; break;
			        case SB_PAGERIGHT: c->val += c->_step; break;
					case SB_LINELEFT:  c->val -= 1; break;
					case SB_LINERIGHT: c->val += 1; break;
                    case SB_THUMBPOSITION: c->val = HIWORD(wParam); break;
                    case SB_BOTTOM:    c->val = c->_min; break;
					case SB_TOP:       c->val = c->_max; break;
                    case SB_THUMBTRACK:       
					case SB_ENDSCROLL: return TRUE;
					default: return FALSE;
                }
				/* an event we recognise, on a control we recognise. wonders never cease. */
				if (c->val < c->_min) c->val = c->_min;
				if (c->val > c->_max) c->val = c->_max;
				redraw_dialog(hDlg);
				return TRUE;
			}
			return FALSE;

		case WM_COMMAND:
			if (p==IDOK) {
				EndDialog(hDlg, TRUE);
				config_save();
				return TRUE;
			} else if (p==IDCANCEL) {
				EndDialog(hDlg, FALSE);
				return TRUE;
			} else if (p==id_defaults) {
				config_item **c = configs();
				for (int i=0; i<n_configdefs; i++) {
					c[i]->val = c[i]->_default;
					/* TODO: Would be nice to reset any T_SELECTs around too. */
				}
				redraw_dialog(hDlg);
				return TRUE;
			}
			/* or maybe it's a dynamic one */
			for (i=0; i<n_configdefs; i++) {
				config_item *c = conf[i];
				if ( (c->type != T_BOOL) && (c->type != T_SELECT) ) 
					continue;
				if (GetDlgItem(hDlg, c->ctlid) != (HWND)lParam)
					continue;
				if (c->type == T_BOOL) {
					if (HIWORD(wParam)==0/*clicked?*/)
						c->val = 1 - c->val;
				} else {
					int t = SendMessage((HWND)lParam, CB_GETCURSEL, 0, 0);
					c->val = c->_kv[t]->val;
				}
				redraw_dialog(hDlg);
				return TRUE;
			}
			return FALSE;

		default: 
			return FALSE;
	}
	// not reached
}  
