/*
 * config.cpp: Configuration management interface
 * Part of Storming Ions (a screensaver).
 * Copyright (C) 2007 W. Ross Younger.
 * See the included file "COPYING" for licensing information.
 */

#include <windows.h>
#include <stdlib.h>
#include <assert.h>

#include "config.h"
#include "base.h"
#include "gfx.h"

// Registry key HKCU/Software/data/APPNAME :
#define APPNAME "crazyscot's storming ions"

/* ****************************************************************************** */

struct _config_context {
	HKEY key;
};

int config_open(config_context *cc) {
	HKEY subkey, ourkey;
	long res;
	DWORD rc;
	config_context rv = 0;

	res=RegOpenKeyEx(HKEY_CURRENT_USER,"Software",0L,KEY_WRITE,&subkey);
	if (res != ERROR_SUCCESS) return 0;

	res=RegCreateKeyEx(subkey,APPNAME,0L,"data",REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, 0L, &ourkey, &rc);
	RegCloseKey(subkey);
	if (res != ERROR_SUCCESS) return 0;

	if(rc==REG_CREATED_NEW_KEY) {
	}

	rv = new struct _config_context();
	rv->key = ourkey;
	*cc = rv;
	return 1;
}

void config_close(config_context cc) {
	if (cc) {
		if (cc->key) RegCloseKey(cc->key);
		delete cc;
	}
}

int config_put(config_context cc, config_item &i) {
	int res;
	assert(cc!=NULL);
	res = RegSetValueEx(cc->key, i.name, 0, REG_DWORD, (LPBYTE)&i.val, sizeof(i.val));
	if (!res) return 1;
	error_msg("Could not write configuration %s to the registry: %d", i.name, GetLastError());
	return 0;
}

int config_read(config_context cc, config_item &i) {
	DWORD res, sz = sizeof(i.val), type;
	assert(cc!=NULL);
	res = RegQueryValueEx(cc->key, i.name, 0, &type, (LPBYTE)&i.val, &sz);
	switch(res) {
		case ERROR_SUCCESS:
			if (type != REG_DWORD) i.val = i._default; /* bibble?! */
			switch (i.type) {
				case T_BOOL: break;
				case T_INT:
				case T_FLOAT:
					if (i.val < i._min) i.val = i._min;
					if (i.val > i._max) i.val = i._max;
					break;
				case T_SELECT: {
					const keyval **kv = i._kv;
					int ok=0;
					while ((*kv)->key) {
						if ((*kv)->val == i.val) ok=1;
						++kv;
					}
					if (!ok) i.val = i._default;
					break;
				}
			}
			break;
		case ERROR_FILE_NOT_FOUND:
		case ERROR_MORE_DATA:
			/* future version? do what we can, NOT an error */
			i.val = i._default;
			break;
		default:
			error_msg("Could not read configuration %s from the registry: %d", i.name, GetLastError());
			return 0;
	}
	return 1;
}

/* ****************************************************************************** */
/* TODO: Remove the runtime string comparisons to read config. */

static config_item **_configs;

unsigned n_configdefs = 0; /* Automatically counted */

static keyval ** draw_kv;

/* pseudo-constructs & loads from registry as necessary... */
config_item **configs() {
	if (!draw_kv) {
		unsigned i=0,n=0;
#define DO(C) do { \
		C(KV("Azimuthal", R_AZIMUTHAL));					\
		C(KV("Conformal", R_CONFORMAL));					\
		C(KV("Conformal (filling)", R_CONFORMAL_ALL));		\
		C(KV("Azimuthal (X-filling)", R_AZIMUTHAL_FILL_X));	\
		C(KV("Azimuthal (Y-filling)", R_AZIMUTHAL_FILL_Y));	\
		C(KV("Debug conformal", R_CONFORMAL_DEBUG));		\
		C(KV("Debug azimuthal", R_AZIMUTHAL_DEBUG));		\
		C(KV(0,0)); /* keyvals must be 0-terminated */		\
		} while (0)
#define X(c) n++
		DO(X);
#undef X
		draw_kv = (keyval**) malloc(n * sizeof(keyval));
#define X(c) draw_kv[i++] = c
		DO(X);
#undef X
#undef DO
		assert(i==n);
	}

	if (!_configs) {
		int i=0, rv;
		n_configdefs=0;

#define DO(C) do { \
		C(CONFIG_SEL("Render mode","Rendering mode", R_AZIMUTHAL, (const keyval**)draw_kv));	\
		C(CONFIG_INT("Redraw interval", "Redraw interval (ms)", 30, 10, 250, 10));				\
		C(CONFIG_INT("Ion size", "Ion size", 3, 1, 4, 1));										\
		C(CONFIG_INT("Number of trails", "Number of trails", 50, 1, 99, 10));					\
		C(CONFIG_INT("Ions per trail", "Ions per trail", 10, 1, 50, 10));						\
		C(CONFIG_INT("Trail gap", "Trail gap", 5, 1, 100, 10));									\
		C(CONFIG_BOOL("Trail gap rand", "Randomise trail gaps?", 0));							\
		C(CONFIG_BOOL("Release bidi", "Bi-directional release?", 0));							\
		C(CONFIG_BOOL("Ordered start", "Ordered start points?", 1));							\
		/* key, help, default, min, max, page, (denominator) */									\
		C(CONFIG_FLOAT("initv scale", "Ion initial vector scale", 10, 1, 30, 1, 10));			\
		C(CONFIG_FLOAT("bh scale", "Black hole gravity", 6, 2, 20, 2, 10000));					\
		C(CONFIG_INT("perturb p","Trail initial position perturbation % limit", 0, 0, 10, 2));	\
		C(CONFIG_INT("perturb v","Trail initial vector perturbation % limit", 0, 0, 500, 2));		\
		} while(0)

#define X(c) n_configdefs++
		DO(X);
#undef X
		_configs = (config_item**)malloc(n_configdefs * sizeof(config_item*));
#define X(c) _configs[i++] = c
		DO(X);
#undef X
#undef DO

		assert(i==n_configdefs);
		/* Still following? :-)
		 * It's not that bad really. Just a trick to dynamically count 
		 * (effectively at compile-time) the number of things in the list.
		 */

		config_context ctx;
		if (1!=(rv=config_open(&ctx))) goto err;
		for (i=0; i<n_configdefs; i++) {
			if (1!=(rv=config_read(ctx, *_configs[i]))) goto err;
		}
err:
		config_close(ctx);
		/* if we had screaming doom loading from the registry, there's little 
		 * we can do about it */
	}

	return _configs;
}


int config_save() {
	int rv, i;
	config_context ctx;
	if (1!=(rv=config_open(&ctx))) goto err;
	for (i=0; i<n_configdefs; i++) {
		if (1!=(rv=config_put(ctx, *_configs[i]))) goto err;
	}
err:
	config_close(ctx);
	return rv;
}

int config_find_index(const char *name) {
	configs();
	for (int i=0; i<n_configdefs; i++) {
		if (0==strcmp(name,_configs[i]->name))
			return i;
	}
	return -1;
}

int config_get(int index) {
	assert(index>=0 && index < n_configdefs);
	return configs()[index]->val;
}

int config_get(const char *name) {
	return config_get(config_find_index(name));
}

double config_get_float(int index) {
	assert(index>=0 && index < n_configdefs);
	unsigned denom = configs()[index]->_denom;
	if (configs()[index]->type != T_FLOAT) denom = 1;
	return configs()[index]->val / (double)denom;
}

double config_get_float(const char *name) {
	return config_get_float(config_find_index(name));
}

/* ****************************************************************************** */