/*
 * gfx.cpp: The main graphics/physics engine.
 * Part of Storming Ions (a screensaver).
 * Copyright (C) 2007 W. Ross Younger.
 * See the included file "COPYING" for licensing information.
 */

#include <windows.h>
#include <wingdi.h>
#include <scrnsave.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <limits.h>

#include "base.h"
#include "gfx.h"
#include "gravity.h"
#include "rng.h"
#include "colours.h"
#include "config.h"

/* Universal constants - no need to make these configurable */
#define EXPER_START_R (1<<14)
#define BH_GRAV (EXPER_START_R*EXPER_START_R)
#define BH_EVENT_HORIZON 1000
#define INITV_PRESCALE (1.0/180)

/* ******************************************************************** */
struct _config {
	unsigned n_trails, ions_per_trail;
	unsigned trail_gap, trail_gap_rand, release_bidi, ion_size, ordered_start;
	long double initv_scale, bh_scale;
	unsigned render_mode /* enum render_mode */;
	unsigned perturb_p, perturb_v;
};

struct _config config;

static void cook_config() {
	config.n_trails = config_get("Number of trails");
	config.ions_per_trail = config_get("Ions per trail");
	config.trail_gap = config_get("Trail gap");
	config.trail_gap_rand = config_get("Trail gap rand");
	config.release_bidi = config_get("Release bidi");
	config.ion_size = config_get("Ion size");
	config.ordered_start = config_get("Ordered start");
	config.initv_scale = config_get_float("initv scale");
	config.bh_scale = config_get_float("bh scale");
	config.render_mode = config_get("Render mode");
	config.perturb_p = config_get("perturb p");
	config.perturb_v = config_get("perturb v");
}

#define MAX_IONS (config.n_trails*config.ions_per_trail)

/* ******************************************************************** */

static void convert_dummy(const coord xin, coord *xout, const coord yin, coord *yout) {
	*xout = *yout = 100;
}

static int scrx, scry;
static void (*master_convert)(const coord xin, coord *xout, const coord yin, coord *yout) = convert_dummy;

/* Conformal mapping.
 * We map the universe, or a sub-section of it, directly onto the screen 
 * area, hence pixels will be stretched unless you have a square display area.
 */
static int conformal_zoom;

#define VIEWSIZE PCT_C(100/conformal_zoom)
#define VIEWMIN (PCT_C(50)-(VIEWSIZE/2))
#define VIEWMAX (PCT_C(50)+(VIEWSIZE/2))

static inline int _C_convert(const coord c, const coord dim) {
	if (c <= VIEWMIN) return -1;
	if (c >= VIEWMAX) return -1;
	return dim*(c - VIEWMIN)/VIEWSIZE;
}

#undef VIEWSIZE
#undef VIEWMIN
#undef VIEWMAX

static void convert_conformal(const coord xin, coord *xout, const coord yin, coord *yout) {
	*xout = _C_convert(xin,scrx);
	*yout = _C_convert(yin,scry);
}

/*
 * Azimuthal mapping.
 * We use the aspect ratio of the screen area to compute a visible region of like ratio,
 * hence keeping pixels square.
 */

static int az_x_min, az_x_max, az_x_size, az_y_min, az_y_max, az_y_size;

static void convert_azimuthal(const coord xin, coord *xout, const coord yin, coord *yout) {
	if ( (xin < az_x_min) || (xin > az_x_max) ) {
		*xout = *yout = -1;
		return;
	}
	if ( (yin < az_y_min) || (yin > az_y_max) ) {
		*xout = *yout = -1;
		return;
	}
	*xout = scrx * (xin - az_x_min) / az_x_size;
	*yout = scry * (yin - az_y_min) / az_y_size;
}

void setup_conversion(RECT &r) {
	scrx = RECT_WIDTH(r);
	scry = RECT_HEIGHT(r);
	double az_scale;

	switch (config.render_mode) {
		case R_CONFORMAL:
			conformal_zoom = 2; break;
		case R_CONFORMAL_ALL:
		case R_CONFORMAL_DEBUG:
			conformal_zoom = 1; break;

		case R_AZIMUTHAL:
			az_y_min = PCT_C(25);
			az_y_max = PCT_C(75);
			az_y_size = az_y_max - az_y_min;
			az_scale = az_y_size / scry;
			az_x_size = az_scale * scrx;
			az_x_min = PCT_C(50) - az_x_size/2;
			az_x_max = az_x_min + az_x_size;
			break;

		case R_AZIMUTHAL_DEBUG:
		case R_AZIMUTHAL_FILL_X:
			az_x_min = PCT_C(0);
			az_x_max = PCT_C(100);
			az_x_size = az_x_max - az_x_min;
			az_y_size = az_x_size*scry/scrx;
			az_y_min = PCT_C(50) - az_y_size/2;
			az_y_max = az_y_min + az_y_size;
			break;
		case R_AZIMUTHAL_FILL_Y:
			az_y_min = PCT_C(0);
			az_y_max = PCT_C(100);
			az_y_size = az_y_max - az_y_min;
			az_x_size = az_y_size*scrx/scry;
			az_x_min = PCT_C(50) - az_x_size/2;
			az_x_max = az_x_min + az_x_size;
			break;

		default:
			error_msg("Unhandled mode %d", config.render_mode);
	}

	switch (config.render_mode) {
		case R_CONFORMAL:
		case R_CONFORMAL_ALL:
		case R_CONFORMAL_DEBUG:
			master_convert = convert_conformal; break;
		case R_AZIMUTHAL:
		case R_AZIMUTHAL_DEBUG:
		case R_AZIMUTHAL_FILL_X:
		case R_AZIMUTHAL_FILL_Y:
			master_convert = convert_azimuthal; break;
		default:
			error_msg("Unhandled mode %d", config.render_mode);
	}
}

class Ion {
private:
	Point _p; /* actual location */
	Point _v; /* velocity vector */
	short _clr; /* colour index */
	unsigned _st; /* no of frames until this ion is live */
	int	  _clamped;
public:
	/* empty cons - constituents created */
	Ion() : _p(), _v(), _clr(0), _st(0), _clamped(0) {} ;
	Ion(Point p, Point v, short clr, unsigned st) : _p(p), _v(v), _clr(clr), _st(st), _clamped(0) { };
	virtual ~Ion(void) { }

	const Point& p() { return _p; }
	const Point& v() { return _v; }
	unsigned plottable() {
		if (_st) {
			_st--;
			return 0;
		}
		if (_clamped)
			return 0;
		return 1;
	}

	/* Tweak this Ion's vector by the influence of a gravitic object.
	 * Newton says: F = G.(m1.m2)/r^2 bilaterally
	 * or, by applying F=ma and cheating and not moving the Body:
	 *		a = G.m/r^2
	 * and as G is a universe constant, and m is assumed not to change, replace G.m by a single constant.
	 * Acceleration, therefore, is a vector DIRECTLY TOWARDS the body, of magnitude GM/r^2.
	 */
	void influence(Body& b) {
		Point& v_to_b = const_cast<Point&>(b.p()) - this->p();
		long long r = v_to_b.mag();
		if (r<BH_EVENT_HORIZON) {
			clamp(); return;
		}
		v_to_b.scale(b.GM() / (r*r));
		v_to_b.scale(config.bh_scale);
		this->_v += v_to_b;
	}
	void influence(Body* b) { influence(*b); }

	void clamp() {
		_p.x = _p.y = MAX_C+1;
		_v.x = _v.y = 0;
		_clamped=1;
	}

	void apply_v() {
		_p += _v;
		/* if we're outside the arena, clamp to stop overflow */
		if ( (_p.x <= MIN_C) || (_p.x >= MAX_C) || (_p.y <= MIN_C) || (_p.y >= MAX_C) )
			clamp();
	}

	void plot(HDC hdc, int mode) {
		int px,py; /* pixel coords to plot */
		if (_clamped) return;

		master_convert(_p.x, &px, _p.y, &py);
#if 0
		switch(config.render_mode) {
			case R_CONFORMAL_DEBUG:
			case R_CONFORMAL_ALL:
			case R_CONFORMAL:
				convert_conformal(_p.x, &px, _p.y, &py);
				break;
			case R_AZIMUTHAL:
			case R_AZIMUTHAL_FILL_X:
			case R_AZIMUTHAL_FILL_Y:
			case R_AZIMUTHAL_DEBUG:
				convert_azimuthal(_p.x, &px, _p.y, &py);
				break;
		}
#endif

		if ( (px>-1) && (py>-1) ) {
			COLORREF c;
			switch (mode) {
				case MODE_BLACK: c = BLACK; break;
				case MODE_WHITE: c = WHITE; break;
				case MODE_PEN:   c = colours[_clr]; break;
			}

			switch (config.ion_size) {
				default:
				case 4:
				SetPixel(hdc,px+2,py,  c);
				SetPixel(hdc,px-2,py,  c);
				SetPixel(hdc,px,  py+2,c);
				SetPixel(hdc,px,  py-2,c);
				case 3:
				SetPixel(hdc,px-1,py-1,c);
				SetPixel(hdc,px-1,py+1,c);
				SetPixel(hdc,px+1,py-1,c);
				SetPixel(hdc,px+1,py+1,c);
				case 2:
				SetPixel(hdc,px+1,py,  c);
				SetPixel(hdc,px-1,py,  c);
				SetPixel(hdc,px,  py+1,c);
				SetPixel(hdc,px,  py-1,c);
				case 1:
				SetPixel(    hdc,px  ,py,  c);
			}
		}
	}
};

/* ******************************************************************** */

/* Run data */
static Body *bh = 0; /* black hole */
static Ion *ions = 0;
static int n_ions = 0;
static int frame;
static Rand rng;


int start_over(HDC hdc, RECT rect)
{
	HGDIOBJ orig;
	PAINTSTRUCT ps;
	int i,j;
	short next_colour = rng.n(gColourCount);

	FillRect (hdc, &rect, (HBRUSH)GetStockObject(BLACK_BRUSH)); 

	/* Where's the black hole going to be? Somewhere within the central quarter-by-area of the screen. */
	delete bh;
	bh = new Body(rng.range(PCT_C(40), PCT_C(60)), rng.range(PCT_C(40), PCT_C(60)), BH_GRAV);

	if ( (config.render_mode == R_CONFORMAL_DEBUG) || (config.render_mode == R_AZIMUTHAL_DEBUG) ) {
		/* ... then show gravity's reach */
		select_colour(hdc,0);
		SelectObject(hdc,(HBRUSH)GetStockObject(BLACK_BRUSH));
		// radius = EXPER_START_R
		int bhx = bh->p().x, bhy = bh->p().y, x1, y1, x2, y2;
		master_convert(bhx-EXPER_START_R, &x1, bhy-EXPER_START_R, &y1);
		master_convert(bhx+EXPER_START_R, &x2, bhy+EXPER_START_R, &y2);
		Ellipse(hdc, x1, y1, x2, y2);
	}

	Point& ppt = bh->p();
	
	if (config.render_mode == R_CONFORMAL_DEBUG) {
		int i,j,x,y;
		convert_conformal(bh->p().x, &x, bh->p().y, &y);
		for (i=-4; i<5; i++)
			for (j=-4; j<5; j++)
				SetPixel(hdc,x+i,y+j,WHITE);
	}

	if(ions) {
		for (i=0; i<n_ions; i++) {
			Ion& ii = ions[i];
			delete &ii;
		}
		delete[] ions;
	}
	ions = new Ion[config.n_trails * config.ions_per_trail];
	n_ions=0;

	for (i=0; i<config.n_trails; i++) {
		// Circular start locus.
		// r = START_R, theta = random.

		double theta;
		if (config.ordered_start)
			theta = (2*3.14159) * i / config.n_trails;
		else
			theta = rng.n(2*314159)/100000.0;
		unsigned r = EXPER_START_R;
		Point trail_base( bh->x() + r*cos(theta), bh->y() + r*sin(theta));

		/* initial trajectory _towards_ the black hole. */
		Point vv = const_cast<Point&>(bh->p()) - trail_base;
		vv.scale(INITV_PRESCALE);
		
		/* And now rotate 90 degrees ... */
		Point v =     Point(-vv.y,vv.x );
		Point vbidi = Point(vv.y, -vv.x);
		if (config.perturb_p) {
			trail_base.x += rng.range(-PCT_C(config.perturb_p), PCT_C(config.perturb_p));
			trail_base.y += rng.range(-PCT_C(config.perturb_p), PCT_C(config.perturb_p));
		}
		v.scale(config.initv_scale);
		if (config.perturb_v) {
			v.scale(1.0 + rng.range(-config.perturb_v,config.perturb_v)/100.0);
		}

		int start_time = 0;

		for (j=0; j< config.ions_per_trail; j++) {
			int bidi = (config.release_bidi && rng.n(2));
			ions[config.ions_per_trail*i+j] = Ion(trail_base,bidi ? vbidi : v,next_colour, start_time);
			next_colour = (next_colour+1) % gColourCount;
			++n_ions;

			if (config.trail_gap_rand)
				start_time += rng.n(config.trail_gap)+1;
			else
				start_time+=config.trail_gap;
		}
	}

	frame=0;
	return 0;
}

int ss_setup(HDC hdc, RECT rect)
{
	int rc;

	cook_config();

	rc=init_colours(MAX_IONS);
	if (rc<0) return rc;

	setup_conversion(rect);

	frame=-1;
	return 0;
}


void ss_frame(HDC hdc, RECT rect)
{
	static int x,y,c;
	int i;

	if (frame<0) start_over(hdc,rect);

	SetROP2(hdc,R2_COPYPEN);

	for (i=0; i<n_ions; i++) {
		Ion *ion = &ions[i];
		if (0==ion->plottable()) continue;
		ion->plot(hdc,MODE_BLACK);
		ion->influence(bh);
		ion->apply_v();
		ion->plot(hdc,MODE_PEN);
	}

	++frame;

	/* Trigger a respawn if there are no live ions left */
	if (!(frame%100)) {

		for (i=0; i<n_ions; i++) {
			if (ions[i].v().x || ions[i].v().y) goto live;
		}
		frame=-1; /* none still running */
live:;
	}
}

void ss_cleanup(void)
{
	cleanup_colours();
}

