/*
 * gravity.h: Basic universe setup - not really gravity, just helper classes for points and gravitic bodies.
 * Part of Storming Ions (a screensaver).
 * Copyright (C) 2007 W. Ross Younger.
 * See the included file "COPYING" for licensing information.
 */

#ifndef GRAVITY_H
#define GRAVITY_H

#include "gfx.h"

/* "Computer, what is the nature of the universe?"
 * "The universe is a two-dimensional Cartesian grid, 2^16 units in both axes." */

/* 
 * In our special universe, we're only using integers to keep the maths nice and fast.
 * As we plot things, grid co-ordinates are translated onto screen co-ordinates, 
 * depending on the rendering mode selected.
 *
 * It might be a sort-of conformal projection distorting the universe to fill the screen 
 * ("whole universe"), or it might be a sort-of azimuthal projection of an area of the
 * universe in which pixels remain square ("normal").
 */
typedef signed int coord;
#define MIN_C 0
#define MAX_C 65535
#define PCT_C(p) ((1+MAX_C)*p/100) /* Absolute percentage co-ordinate, e.g. 50% = centre */

class Point {
public:
	coord x,y;
	Point() { Point(0,0); };
	Point(coord xx, coord yy) : x(xx), y(yy) {};

	operator Point() { return *this; }
	
	Point operator+(const Point& other) {
		Point t = Point(x,y);
		t += other;
		return t;
	}
	Point operator-(const Point& other) {
		Point t = Point(x,y);
		t -= other;
		return t;
	}
	Point& operator+=(const Point& other) {
		x += other.x; y += other.y;
		return *this;
	}
	Point& operator-=(const Point& other) {
		x -= other.x; y -= other.y;
		return *this;
	}
	const double mag() {
		return sqrt( (double) this->x*this->x + this->y*this->y );
	}
	void scale(double factor) {
		x *= factor;
		y *= factor;
	}
	Point* newScale(double factor) {
		return new Point(x*factor, y*factor);
	}
};




/* Class representing a static body. */
class Body { /* TODO: Make this inherit from Point. */
private:
	Point _p;
	unsigned _GM; /* gravitic constant TIMES mass */
public:
	Body(coord x, coord y, unsigned GM) : _p(Point(x,y)), _GM(GM) {};
	Body(Point pp, unsigned GM) : _p(pp), _GM(GM) {};
	Body() : _p(), _GM(1e9) {} ;
	virtual ~Body(void) { }

	Point& p() { return _p; }
	const unsigned GM() { return _GM; }
	coord x() { return _p.x; }
	coord y() { return _p.y; }
};

#endif
