/*
 * colours.cpp: Random number generation.
 * Part of Storming Ions (a screensaver).
 * Copyright (C) 2007 W. Ross Younger.
 * See the included file "COPYING" for licensing information.
 */

/* stdlib's rand() is pretty pants.
 * I want:
   * More entropy from a single shot
   * Helper wrapper(s) for _unbiased_ sampling
 *
 * So I googled, and found a generator and helper code with explicitly-PD
 * source.
 *   http://www.helsbreth.org/random/shuffle.html
 * - downloaded on 9 Oct 2007. C++ised, and wincrypt seeding lashup, by me.
 */

#include <windows.h>
#include <wincrypt.h>

#include "rng.h"

/* ----------------------------------------------------------------------

COMBO

  a simple but very good combination generator.
  It combines, by addition mod 2^32,
            x(n)=x(n-1)*x(n-2) mod 2^32
        and
            y(n)=30903*y(n-1) + carry mod 2^16
 The period of the first is 3*2^29, on odd integers, and the
 period of the second, a multiply-with-carry generator, is
 30903*2^15-1=1012629503, so the period of combo exceeds 2^60.
 This generator is simple to program in Fortran or C and quite
 fast.

*/

/*
 Simple combo, period > 2^60.5
 x(n)=x(n-1)*x(n-2) mod 2^32 added to
 period of x(n)=x(n-1)*x(n-2) is 3*2^29 if both seeds are
    odd, and one is +or-3 mod 8.
 easy to ensure: replace seed x with 8*seed+3, and y with 2*seed+1
*/

static unsigned long winrand() {
	HCRYPTPROV cProvider;
	unsigned long rv;
	if (!CryptAcquireContext(&cProvider, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) throw GetLastError();
	if (!CryptGenRandom(cProvider, sizeof(rv), (BYTE*)&rv)) throw GetLastError();
	CryptReleaseContext(cProvider,0);
	return rv;
}

void Rand::init(unsigned long seed) {
	combo_x = seed*8+3;
	combo_y = seed*2+1;
	combo_z = seed|1;
	combo_v = 0;
}

/* Implicit seeding */
Rand::Rand() : MAX(4294967295UL) { init(winrand()); }

/* Explicit seeding */
Rand::Rand(unsigned long seed) : MAX(4294967295UL) { init(seed); }

unsigned long Rand::next() {
    combo_v = combo_x * combo_y;
    combo_x = combo_y;
    combo_y = combo_v;
    combo_z = (combo_z & 65535) * 30903 + (combo_z >> 16);
    return combo_y + combo_z;
}


/* ----------------------------------------------------------------------
  range
  
  returns a number between 0 and N-1
  without any bias.

*/

unsigned Rand::n(unsigned n) {
   unsigned long div;
   int r;

   div = this->MAX / n;

   do {
      r = this->next() / div;
   } while (r >= n);

   return r;
}
/* -------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   range2

   returns an integer between low and high inclusive

*/

int Rand::range(int low, int high) {
   return low + this->n(high-low+1);
}